@@ -1,7 +1,7 @@
 /*
  * The initial "port" of dwm to curses was done by
  *
- * © 2007-2014 Marc André Tanner <mat at brain-dump dot org>
+ * © 2007-2016 Marc André Tanner <mat at brain-dump dot org>
  *
  * It is highly inspired by the original X11 dwm and
  * reuses some code of it which is mostly
@@ -10,8 +10,11 @@
  *
  * See LICENSE for details.
  */
-
-#define _GNU_SOURCE
+#include <stdlib.h>
+#include <unistd.h>
+#include <stdint.h>
+#include <wchar.h>
+#include <sys/select.h>
 #include <sys/stat.h>
 #include <sys/ioctl.h>
 #include <sys/wait.h>
@@ -27,7 +30,8 @@
 #include <unistd.h>
 #include <stdbool.h>
 #include <errno.h>
-#ifdef __CYGWIN__
+#include <pwd.h>
+#if defined __CYGWIN__ || defined __sun
 # include <termios.h>
 #endif
 #include "vt.h"
@@ -42,10 +46,11 @@ int ESCDELAY;
 
 typedef struct {
 	float mfact;
+	unsigned int nmaster;
 	int history;
 	int w;
 	int h;
-	bool need_resize;
+	volatile sig_atomic_t need_resize;
 } Screen;
 
 typedef struct {
@@ -57,11 +62,13 @@ typedef struct Client Client;
 struct Client {
 	WINDOW *window;
 	Vt *term;
+	Vt *editor, *app;
+	int editor_fds[2];
+	volatile sig_atomic_t editor_died;
 	const char *cmd;
 	char title[255];
 	int order;
 	pid_t pid;
-	int pty;
 	unsigned short int id;
 	unsigned short int x;
 	unsigned short int y;
@@ -69,16 +76,26 @@ struct Client {
 	unsigned short int h;
 	bool has_title_line;
 	bool minimized;
-	bool died;
+	bool urgent;
+	volatile sig_atomic_t died;
 	Client *next;
 	Client *prev;
+	Client *snext;
+	unsigned int tags;
 };
 
 typedef struct {
-	const char *title;
-	unsigned attrs;
 	short fg;
 	short bg;
+	short fg256;
+	short bg256;
+	short pair;
+} Color;
+
+typedef struct {
+	const char *title;
+	attr_t attrs;
+	Color *color;
 } ColorRule;
 
 #define ALT(k)      ((k) + (161 - 'a'))
@@ -94,16 +111,17 @@ typedef struct {
 
 typedef struct {
 	void (*cmd)(const char *args[]);
-	/* needed to avoid an error about initialization
-	 * of nested flexible array members */
 	const char *args[MAX_ARGS];
 } Action;
 
+#define MAX_KEYS 3
+
+typedef unsigned int KeyCombo[MAX_KEYS];
+
 typedef struct {
-	unsigned int mod;
-	unsigned int code;
+	KeyCombo keys;
 	Action action;
-} Key;
+} KeyBinding;
 
 typedef struct {
 	mmask_t mask;
@@ -116,11 +134,11 @@ typedef struct {
 } Cmd;
 
 enum { BAR_TOP, BAR_BOTTOM, BAR_OFF };
-enum { ALIGN_LEFT, ALIGN_RIGHT };
 
 typedef struct {
 	int fd;
-	int pos;
+	int pos, lastpos;
+	bool autohide;
 	unsigned short int h;
 	unsigned short int y;
 	char text[512];
@@ -133,9 +151,23 @@ typedef struct {
 	unsigned short int id;
 } CmdFifo;
 
-#define countof(arr) (sizeof(arr) / sizeof((arr)[0]))
-#define sstrlen(str) (sizeof(str) - 1)
-#define max(x, y) ((x) > (y) ? (x) : (y))
+typedef struct {
+	char *data;
+	size_t len;
+	size_t size;
+} Register;
+
+typedef struct {
+	char *name;
+	const char *argv[4];
+	bool filter;
+	bool color;
+} Editor;
+
+#define LENGTH(arr) (sizeof(arr) / sizeof((arr)[0]))
+#define MAX(x, y)   ((x) > (y) ? (x) : (y))
+#define MIN(x, y)   ((x) < (y) ? (x) : (y))
+#define TAGMASK     ((1 << LENGTH(tags)) - 1)
 
 #ifdef NDEBUG
  #define debug(format, args...)
@@ -159,13 +191,19 @@ static void redraw(const char *args[]);
 static void scrollback(const char *args[]);
 static void send(const char *args[]);
 static void setlayout(const char *args[]);
+static void incnmaster(const char *args[]);
 static void setmfact(const char *args[]);
 static void startup(const char *args[]);
+static void tag(const char *args[]);
 static void togglebar(const char *args[]);
-static void togglebell(const char *key[]);
+static void togglebarpos(const char *args[]);
 static void toggleminimize(const char *args[]);
 static void togglemouse(const char *args[]);
 static void togglerunall(const char *args[]);
+static void toggletag(const char *args[]);
+static void toggleview(const char *args[]);
+static void viewprevtag(const char *args[]);
+static void view(const char *args[]);
 static void zoom(const char *args[]);
 
 /* commands for use by mouse bindings */
@@ -175,27 +213,31 @@ static void mouse_minimize(const char *args[]);
 static void mouse_zoom(const char *args[]);
 
 /* functions and variables available to layouts via config.h */
+static Client* nextvisible(Client *c);
+static void focus(Client *c);
 static void resize(Client *c, int x, int y, int w, int h);
 extern Screen screen;
 static unsigned int waw, wah, wax, way;
 static Client *clients = NULL;
 static char *title;
-#define COLOR(fg, bg) COLOR_PAIR(vt_color_reserve(fg, bg))
-#define NOMOD ERR
 
 #include "config.h"
 
 /* global variables */
-Screen screen = { MFACT, SCROLL_HISTORY };
+static const char *dvtm_name = "dvtm";
+Screen screen = { .mfact = MFACT, .nmaster = NMASTER, .history = SCROLL_HISTORY };
+static Client *stack = NULL;
 static Client *sel = NULL;
 static Client *lastsel = NULL;
 static Client *msel = NULL;
+static unsigned int seltags;
+static unsigned int tagset[2] = { 1, 1 };
 static bool mouse_events_enabled = ENABLE_MOUSE;
 static Layout *layout = layouts;
-static StatusBar bar = { -1, BAR_POS, 1 };
-static CmdFifo cmdfifo = { -1 };
+static StatusBar bar = { .fd = -1, .lastpos = BAR_POS, .pos = BAR_POS, .autohide = BAR_AUTOHIDE, .h = 1 };
+static CmdFifo cmdfifo = { .fd = -1 };
 static const char *shell;
-static char *copybuf;
+static Register copyreg;
 static volatile sig_atomic_t running = true;
 static bool runinall = false;
 
@@ -221,59 +263,141 @@ isarrange(void (*func)()) {
 	return func == layout->arrange;
 }
 
+static bool
+isvisible(Client *c) {
+	return c->tags & tagset[seltags];
+}
+
 static bool
 is_content_visible(Client *c) {
 	if (!c)
 		return false;
 	if (isarrange(fullscreen))
 		return sel == c;
-	return !c->minimized;
+	return isvisible(c) && !c->minimized;
+}
+
+static Client*
+nextvisible(Client *c) {
+	for (; c && !isvisible(c); c = c->next);
+	return c;
 }
 
 static void
-drawbar() {
-	wchar_t wbuf[sizeof bar.text];
-	int x, y, w, maxwidth = screen.w - 2;
+updatebarpos(void) {
+	bar.y = 0;
+	wax = 0;
+	way = 0;
+	wah = screen.h;
+	waw = screen.w;
+	if (bar.pos == BAR_TOP) {
+		wah -= bar.h;
+		way += bar.h;
+	} else if (bar.pos == BAR_BOTTOM) {
+		wah -= bar.h;
+		bar.y = wah;
+	}
+}
+
+static void
+hidebar(void) {
+	if (bar.pos != BAR_OFF) {
+		bar.lastpos = bar.pos;
+		bar.pos = BAR_OFF;
+	}
+}
+
+static void
+showbar(void) {
+	if (bar.pos == BAR_OFF)
+		bar.pos = bar.lastpos;
+}
+
+static void
+drawbar(void) {
+	int sx, sy, x, y, width;
+	unsigned int occupied = 0, urgent = 0;
 	if (bar.pos == BAR_OFF)
 		return;
-	getyx(stdscr, y, x);
+
+	for (Client *c = clients; c; c = c->next) {
+		occupied |= c->tags;
+		if (c->urgent)
+			urgent |= c->tags;
+	}
+
+	getyx(stdscr, sy, sx);
 	attrset(BAR_ATTR);
-	mvaddch(bar.y, 0, '[');
-	if (mbstowcs(wbuf, bar.text, sizeof bar.text) == (size_t)-1)
-		return;
-	if ((w = wcswidth(wbuf, maxwidth)) == -1)
-		return;
-	if (BAR_ALIGN == ALIGN_RIGHT) {
-		for (int i = 0; i + w < maxwidth; i++)
-			addch(' ');
+	move(bar.y, 0);
+
+	for (unsigned int i = 0; i < LENGTH(tags); i++){
+		if (tagset[seltags] & (1 << i))
+			attrset(TAG_SEL);
+		else if (urgent & (1 << i))
+			attrset(TAG_URGENT);
+		else if (occupied & (1 << i))
+			attrset(TAG_OCCUPIED);
+		else
+			attrset(TAG_NORMAL);
+		printw(TAG_SYMBOL, tags[i]);
 	}
-	addstr(bar.text);
-	if (BAR_ALIGN == ALIGN_LEFT) {
-		for (; w < maxwidth; w++)
+
+	attrset(TAG_NORMAL);
+	addstr(layout->symbol);
+
+	getyx(stdscr, y, x);
+	(void)y;
+	int maxwidth = screen.w - x - 2;
+
+	addch(BAR_BEGIN);
+	attrset(BAR_ATTR);
+
+	wchar_t wbuf[sizeof bar.text];
+	size_t numchars = mbstowcs(wbuf, bar.text, sizeof bar.text);
+
+	if (numchars != (size_t)-1 && (width = wcswidth(wbuf, maxwidth)) != -1) {
+		int pos;
+		for (pos = 0; pos + width < maxwidth; pos++)
 			addch(' ');
+
+		for (size_t i = 0; i < numchars; i++) {
+			pos += wcwidth(wbuf[i]);
+			if (pos > maxwidth)
+				break;
+			addnwstr(wbuf+i, 1);
+		}
+
+		clrtoeol();
 	}
-	mvaddch(bar.y, screen.w - 1, ']');
+
+	attrset(TAG_NORMAL);
+	mvaddch(bar.y, screen.w - 1, BAR_END);
 	attrset(NORMAL_ATTR);
-	move(y, x);
+	move(sy, sx);
 	wnoutrefresh(stdscr);
 }
 
 static int
-show_border() {
-	return (bar.fd != -1 && bar.pos != BAR_OFF) || (clients && clients->next);
+show_border(void) {
+	return (bar.pos != BAR_OFF) || (clients && clients->next);
 }
 
 static void
 draw_border(Client *c) {
+	char t = '\0';
+	int x, y, maxlen, attrs = NORMAL_ATTR;
+
 	if (!show_border())
 		return;
-	char t = '\0';
-	int x, y, maxlen;
+	if (sel == c || (runinall && !c->minimized))
+		attrs = SELECTED_ATTR;
+	if (sel != c && c->urgent)
+		attrs = URGENT_ATTR;
 
-	wattrset(c->window, (sel == c || (runinall && !c->minimized)) ? SELECTED_ATTR : NORMAL_ATTR);
+	wattrset(c->window, attrs);
 	getyx(c->window, y, x);
 	mvwhline(c->window, 0, 0, ACS_HLINE, c->w);
-	maxlen = c->w - (2 + sstrlen(TITLE) - sstrlen("%s%sd")  + sstrlen(SEPARATOR) + 2);
+	maxlen = c->w - 10;
 	if (maxlen < 0)
 		maxlen = 0;
 	if ((size_t)maxlen < sizeof(c->title)) {
@@ -281,9 +405,9 @@ draw_border(Client *c) {
 		c->title[maxlen] = '\0';
 	}
 
-	mvwprintw(c->window, 0, 2, TITLE,
+	mvwprintw(c->window, 0, 2, "[%s%s#%d]",
 	          *c->title ? c->title : "",
-	          *c->title ? SEPARATOR : "",
+	          *c->title ? " | " : "",
 	          c->order);
 	if (t)
 		c->title[maxlen] = t;
@@ -307,9 +431,18 @@ draw(Client *c) {
 }
 
 static void
-draw_all() {
+draw_all(void) {
+	if (!nextvisible(clients)) {
+		sel = NULL;
+		curs_set(0);
+		erase();
+		drawbar();
+		doupdate();
+		return;
+	}
+
 	if (!isarrange(fullscreen)) {
-		for (Client *c = clients; c; c = c->next) {
+		for (Client *c = nextvisible(clients); c; c = nextvisible(c->next)) {
 			if (c == sel)
 				continue;
 			draw(c);
@@ -324,12 +457,38 @@ draw_all() {
 }
 
 static void
-arrange() {
+arrange(void) {
+	unsigned int m = 0, n = 0;
+	for (Client *c = nextvisible(clients); c; c = nextvisible(c->next)) {
+		c->order = ++n;
+		if (c->minimized)
+			m++;
+	}
 	erase();
-	drawbar();
 	attrset(NORMAL_ATTR);
+	if (bar.fd == -1 && bar.autohide) {
+		if ((!clients || !clients->next) && n == 1)
+			hidebar();
+		else
+			showbar();
+		updatebarpos();
+	}
+	if (m && !isarrange(fullscreen))
+		wah--;
 	layout->arrange();
+	if (m && !isarrange(fullscreen)) {
+		unsigned int i = 0, nw = waw / m, nx = wax;
+		for (Client *c = nextvisible(clients); c; c = nextvisible(c->next)) {
+			if (c->minimized) {
+				resize(c, nx, way+wah, ++i == m ? waw - nx : nw, 1);
+				nx += nw;
+			}
+		}
+		wah++;
+	}
+	focus(NULL);
 	wnoutrefresh(stdscr);
+	drawbar();
 	draw_all();
 }
 
@@ -340,7 +499,7 @@ attach(Client *c) {
 	c->next = clients;
 	c->prev = NULL;
 	clients = c;
-	for (int o = 1; c; c = c->next, o++)
+	for (int o = 1; c; c = nextvisible(c->next), o++)
 		c->order = o;
 }
 
@@ -357,11 +516,17 @@ attachafter(Client *c, Client *a) { /* attach c after a */
 		c->next = a->next;
 		c->prev = a;
 		a->next = c;
-		for (int o = a->order; c; c = c->next)
+		for (int o = a->order; c; c = nextvisible(c->next))
 			c->order = ++o;
 	}
 }
 
+static void
+attachstack(Client *c) {
+	c->snext = stack;
+	stack = c;
+}
+
 static void
 detach(Client *c) {
 	Client *d;
@@ -369,7 +534,7 @@ detach(Client *c) {
 		c->prev->next = c->next;
 	if (c->next) {
 		c->next->prev = c->prev;
-		for (d = c->next; d; d = d->next)
+		for (d = nextvisible(c->next); d; d = nextvisible(d->next))
 			--d->order;
 	}
 	if (c == clients)
@@ -386,66 +551,84 @@ settitle(Client *c) {
 		printf("\033]0;%s\007", t);
 }
 
+static void
+detachstack(Client *c) {
+	Client **tc;
+	for (tc=&stack; *tc && *tc != c; tc=&(*tc)->snext);
+	*tc = c->snext;
+}
+
 static void
 focus(Client *c) {
-	Client *tmp = sel;
+	if (!c)
+		for (c = stack; c && !isvisible(c); c = c->snext);
 	if (sel == c)
 		return;
 	lastsel = sel;
 	sel = c;
-	settitle(c);
-	if (tmp && !isarrange(fullscreen)) {
-		draw_border(tmp);
-		wnoutrefresh(tmp->window);
+	if (lastsel) {
+		lastsel->urgent = false;
+		if (!isarrange(fullscreen)) {
+			draw_border(lastsel);
+			wnoutrefresh(lastsel->window);
+		}
 	}
-	if (isarrange(fullscreen)) {
-		draw(c);
-	} else {
-		draw_border(c);
-		wnoutrefresh(c->window);
+
+	if (c) {
+		detachstack(c);
+		attachstack(c);
+		settitle(c);
+		c->urgent = false;
+		if (isarrange(fullscreen)) {
+			draw(c);
+		} else {
+			draw_border(c);
+			wnoutrefresh(c->window);
+		}
 	}
-	curs_set(!c->minimized && vt_cursor(c->term));
+	curs_set(c && !c->minimized && vt_cursor_visible(c->term));
 }
 
 static void
 applycolorrules(Client *c) {
 	const ColorRule *r = colorrules;
-	short fg = r->fg, bg = r->bg;
-	unsigned attrs = r->attrs;
+	short fg = r->color->fg, bg = r->color->bg;
+	attr_t attrs = r->attrs;
 
-	for (unsigned int i = 1; i < countof(colorrules); i++) {
+	for (unsigned int i = 1; i < LENGTH(colorrules); i++) {
 		r = &colorrules[i];
 		if (strstr(c->title, r->title)) {
 			attrs = r->attrs;
-			fg = r->fg;
-			bg = r->bg;
+			fg = r->color->fg;
+			bg = r->color->bg;
 			break;
 		}
 	}
 
-	vt_set_default_colors(c->term, attrs, fg, bg);
+	vt_default_colors_set(c->term, attrs, fg, bg);
 }
 
 static void
-term_event_handler(Vt *term, int event, void *event_data) {
-	Client *c = (Client *)vt_get_data(term);
-	switch (event) {
-	case VT_EVENT_TITLE:
-		if (event_data)
-			strncpy(c->title, event_data, sizeof(c->title) - 1);
-		c->title[event_data ? sizeof(c->title) - 1 : 0] = '\0';
-		settitle(c);
-		if (!isarrange(fullscreen) || sel == c)
-			draw_border(c);
-		applycolorrules(c);
-		break;
-	case VT_EVENT_COPY_TEXT:
-		if (event_data) {
-			free(copybuf);
-			copybuf = event_data;
-		}
-		break;
-	}
+term_title_handler(Vt *term, const char *title) {
+	Client *c = (Client *)vt_data_get(term);
+	if (title)
+		strncpy(c->title, title, sizeof(c->title) - 1);
+	c->title[title ? sizeof(c->title) - 1 : 0] = '\0';
+	settitle(c);
+	if (!isarrange(fullscreen) || sel == c)
+		draw_border(c);
+	applycolorrules(c);
+}
+
+static void
+term_urgent_handler(Vt *term) {
+	Client *c = (Client *)vt_data_get(term);
+	c->urgent = true;
+	printf("\a");
+	fflush(stdout);
+	drawbar();
+	if (!isarrange(fullscreen) && sel != c && isvisible(c))
+		draw_border(c);
 }
 
 static void
@@ -476,7 +659,9 @@ resize_client(Client *c, int w, int h) {
 	}
 	if (resize_window || c->has_title_line != has_title_line) {
 		c->has_title_line = has_title_line;
-		vt_resize(c->term, h - has_title_line, w);
+		vt_resize(c->app, h - has_title_line, w);
+		if (c->editor)
+			vt_resize(c->editor, h - has_title_line, w);
 	}
 }
 
@@ -486,22 +671,13 @@ resize(Client *c, int x, int y, int w, int h) {
 	move_client(c, x, y);
 }
 
-static Client*
-get_client_by_pid(pid_t pid) {
-	for (Client *c = clients; c; c = c->next) {
-		if (c->pid == pid)
-			return c;
-	}
-	return NULL;
-}
-
 static Client*
 get_client_by_coord(unsigned int x, unsigned int y) {
 	if (y < way || y >= wah)
 		return NULL;
 	if (isarrange(fullscreen))
 		return sel;
-	for (Client *c = clients; c; c = c->next) {
+	for (Client *c = nextvisible(clients); c; c = nextvisible(c->next)) {
 		if (x >= c->x && x < c->x + c->w && y >= c->y && y < c->y + c->h) {
 			debug("mouse event, x: %d y: %d client: %d\n", x, y, c->order);
 			return c;
@@ -525,10 +701,19 @@ sigchld_handler(int sig) {
 			eprint("waitpid: %s\n", strerror(errno));
 			break;
 		}
+
 		debug("child with pid %d died\n", pid);
-		Client *c = get_client_by_pid(pid);
-		if (c)
-			c->died = true;
+
+		for (Client *c = clients; c; c = c->next) {
+			if (c->pid == pid) {
+				c->died = true;
+				break;
+			}
+			if (c->editor && vt_pid_get(c->editor) == pid) {
+				c->editor_died = true;
+				break;
+			}
+		}
 	}
 
 	errno = errsv;
@@ -545,24 +730,7 @@ sigterm_handler(int sig) {
 }
 
 static void
-updatebarpos(void) {
-	bar.y = 0;
-	wax = 0;
-	way = 0;
-	wah = screen.h;
-	if (bar.fd == -1)
-		return;
-	if (bar.pos == BAR_TOP) {
-		wah -= bar.h;
-		way += bar.h;
-	} else if (bar.pos == BAR_BOTTOM) {
-		wah -= bar.h;
-		bar.y = wah;
-	}
-}
-
-static void
-resize_screen() {
+resize_screen(void) {
 	struct winsize ws;
 
 	if (ioctl(0, TIOCGWINSZ, &ws) == -1) {
@@ -576,32 +744,94 @@ resize_screen() {
 
 	resizeterm(screen.h, screen.w);
 	wresize(stdscr, screen.h, screen.w);
-
-	waw = screen.w;
-	wah = screen.h;
 	updatebarpos();
 	clear();
 	arrange();
 }
 
-static bool
-is_modifier(unsigned int mod) {
-	for (unsigned int i = 0; i < countof(keys); i++) {
-		if (keys[i].mod == mod)
-			return true;
-	}
-	return false;
-}
-
-static Key*
-keybinding(unsigned int mod, unsigned int code) {
-	for (unsigned int i = 0; i < countof(keys); i++) {
-		if (keys[i].mod == mod && keys[i].code == code)
-			return &keys[i];
+static KeyBinding*
+keybinding(KeyCombo keys) {
+	unsigned int keycount = 0;
+	while (keycount < MAX_KEYS && keys[keycount])
+		keycount++;
+	for (unsigned int b = 0; b < LENGTH(bindings); b++) {
+		for (unsigned int k = 0; k < keycount; k++) {
+			if (keys[k] != bindings[b].keys[k])
+				break;
+			if (k == keycount - 1)
+				return &bindings[b];
+		}
 	}
 	return NULL;
 }
 
+static unsigned int
+bitoftag(const char *tag) {
+	unsigned int i;
+	for (i = 0; (i < LENGTH(tags)) && (tags[i] != tag); i++);
+	return (i < LENGTH(tags)) ? (1 << i) : ~0;
+}
+
+static void
+tagschanged() {
+	bool allminimized = true;
+	for (Client *c = nextvisible(clients); c; c = nextvisible(c->next)) {
+		if (!c->minimized) {
+			allminimized = false;
+			break;
+		}
+	}
+	if (allminimized && nextvisible(clients)) {
+		focus(NULL);
+		toggleminimize(NULL);
+	}
+	arrange();
+}
+
+static void
+tag(const char *args[]) {
+	if (!sel)
+		return;
+	sel->tags = bitoftag(args[0]) & TAGMASK;
+	tagschanged();
+}
+
+static void
+toggletag(const char *args[]) {
+	if (!sel)
+		return;
+	unsigned int newtags = sel->tags ^ (bitoftag(args[0]) & TAGMASK);
+	if (newtags) {
+		sel->tags = newtags;
+		tagschanged();
+	}
+}
+
+static void
+toggleview(const char *args[]) {
+	unsigned int newtagset = tagset[seltags] ^ (bitoftag(args[0]) & TAGMASK);
+	if (newtagset) {
+		tagset[seltags] = newtagset;
+		tagschanged();
+	}
+}
+
+static void
+view(const char *args[]) {
+	unsigned int newtagset = bitoftag(args[0]) & TAGMASK;
+	if (tagset[seltags] != newtagset && newtagset) {
+		seltags ^= 1; /* toggle sel tagset */
+		tagset[seltags] = newtagset;
+		tagschanged();
+	}
+}
+
+static void
+viewprevtag(const char *args[]) {
+	seltags ^= 1;
+	tagschanged();
+}
+
 static void
 keypress(int code) {
 	unsigned int len = 1;
@@ -615,8 +845,9 @@ keypress(int code) {
 		nodelay(stdscr, FALSE);
 	}
 
-	for (Client *c = runinall ? clients : sel; c; c = c->next) {
+	for (Client *c = runinall ? nextvisible(clients) : sel; c; c = nextvisible(c->next)) {
 		if (is_content_visible(c)) {
+			c->urgent = false;
 			if (code == '\e')
 				vt_write(c->term, buf, len);
 			else
@@ -628,32 +859,64 @@ keypress(int code) {
 }
 
 static void
-mouse_setup() {
+mouse_setup(void) {
 #ifdef CONFIG_MOUSE
 	mmask_t mask = 0;
 
 	if (mouse_events_enabled) {
 		mask = BUTTON1_CLICKED | BUTTON2_CLICKED;
-		for (unsigned int i = 0; i < countof(buttons); i++)
+		for (unsigned int i = 0; i < LENGTH(buttons); i++)
 			mask |= buttons[i].mask;
 	}
 	mousemask(mask, NULL);
 #endif /* CONFIG_MOUSE */
 }
 
+static bool
+checkshell(const char *shell) {
+	if (shell == NULL || *shell == '\0' || *shell != '/')
+		return false;
+	if (!strcmp(strrchr(shell, '/')+1, dvtm_name))
+		return false;
+	if (access(shell, X_OK))
+		return false;
+	return true;
+}
+
+static const char *
+getshell(void) {
+	const char *shell = getenv("SHELL");
+	struct passwd *pw;
+
+	if (checkshell(shell))
+		return shell;
+	if ((pw = getpwuid(getuid())) && checkshell(pw->pw_shell))
+		return pw->pw_shell;
+	return "/bin/sh";
+}
+
 static void
-setup() {
-	if (!(shell = getenv("SHELL")))
-		shell = "/bin/sh";
+setup(void) {
+	shell = getshell();
 	setlocale(LC_CTYPE, "");
 	initscr();
 	start_color();
 	noecho();
+	nonl();
 	keypad(stdscr, TRUE);
 	mouse_setup();
 	raw();
 	vt_init();
-	vt_set_keytable(keytable, countof(keytable));
+	vt_keytable_set(keytable, LENGTH(keytable));
+	for (unsigned int i = 0; i < LENGTH(colors); i++) {
+		if (COLORS == 256) {
+			if (colors[i].fg256)
+				colors[i].fg = colors[i].fg256;
+			if (colors[i].bg256)
+				colors[i].bg = colors[i].bg256;
+		}
+		colors[i].pair = vt_color_reserve(colors[i].fg, colors[i].bg);
+	}
 	resize_screen();
 	struct sigaction sa;
 	sa.sa_flags = 0;
@@ -664,6 +927,8 @@ setup() {
 	sigaction(SIGCHLD, &sa, NULL);
 	sa.sa_handler = sigterm_handler;
 	sigaction(SIGTERM, &sa, NULL);
+	sa.sa_handler = SIG_IGN;
+	sigaction(SIGPIPE, &sa, NULL);
 }
 
 static void
@@ -671,9 +936,11 @@ destroy(Client *c) {
 	if (sel == c)
 		focusnextnm(NULL);
 	detach(c);
+	detachstack(c);
 	if (sel == c) {
-		if (clients) {
-			focus(clients);
+		Client *next = nextvisible(clients);
+		if (next) {
+			focus(next);
 			toggleminimize(NULL);
 		} else {
 			sel = NULL;
@@ -685,7 +952,7 @@ destroy(Client *c) {
 	wnoutrefresh(c->window);
 	vt_destroy(c->term);
 	delwin(c->window);
-	if (!clients && countof(actions)) {
+	if (!clients && LENGTH(actions)) {
 		if (!strcmp(c->cmd, shell))
 			quit(NULL);
 		else
@@ -696,12 +963,12 @@ destroy(Client *c) {
 }
 
 static void
-cleanup() {
+cleanup(void) {
 	while (clients)
 		destroy(clients);
 	vt_shutdown();
 	endwin();
-	free(copybuf);
+	free(copyreg.data);
 	if (bar.fd > 0)
 		close(bar.fd);
 	if (bar.file)
@@ -720,53 +987,54 @@ static char *getcwd_by_pid(Client *c) {
 	return realpath(buf, NULL);
 }
 
-/* commands for use by keybindings */
 static void
 create(const char *args[]) {
-	Client *c = calloc(1, sizeof(Client));
-	if (!c)
-		return;
-	const char *cmd = (args && args[0]) ? args[0] : shell;
-	const char *pargs[] = { "/bin/sh", "-c", cmd, NULL };
-	c->id = ++cmdfifo.id;
+	const char *pargs[4] = { shell, NULL };
 	char buf[8], *cwd = NULL;
-	snprintf(buf, sizeof buf, "%d", c->id);
 	const char *env[] = {
-		"DVTM", VERSION,
 		"DVTM_WINDOW_ID", buf,
 		NULL
 	};
 
+	if (args && args[0]) {
+		pargs[1] = "-c";
+		pargs[2] = args[0];
+		pargs[3] = NULL;
+	}
+	Client *c = calloc(1, sizeof(Client));
+	if (!c)
+		return;
+	c->tags = tagset[seltags];
+	c->id = ++cmdfifo.id;
+	snprintf(buf, sizeof buf, "%d", c->id);
+
 	if (!(c->window = newwin(wah, waw, way, wax))) {
 		free(c);
 		return;
 	}
 
-	c->has_title_line = show_border();
-	if (!(c->term = vt_create(screen.h - c->has_title_line, screen.w, screen.history))) {
+	c->term = c->app = vt_create(screen.h, screen.w, screen.history);
+	if (!c->term) {
 		delwin(c->window);
 		free(c);
 		return;
 	}
 
-	c->cmd = cmd;
+	c->cmd = (args && args[0]) ? args[0] : shell;
 	if (args && args[1]) {
 		strncpy(c->title, args[1], sizeof(c->title) - 1);
 		c->title[sizeof(c->title) - 1] = '\0';
 	}
 	if (args && args[2])
 		cwd = !strcmp(args[2], "$CWD") ? getcwd_by_pid(sel) : (char*)args[2];
-	c->pid = vt_forkpty(c->term, "/bin/sh", pargs, cwd, env, &c->pty);
+	c->pid = vt_forkpty(c->term, shell, pargs, cwd, env, NULL, NULL);
 	if (args && args[2] && !strcmp(args[2], "$CWD"))
 		free(cwd);
-	vt_set_data(c->term, c);
-	vt_set_event_handler(c->term, term_event_handler);
-	c->w = screen.w;
-	c->h = screen.h;
+	vt_data_set(c->term, c);
+	vt_title_handler_set(c->term, term_title_handler);
+	vt_urgent_handler_set(c->term, term_urgent_handler);
 	c->x = wax;
 	c->y = way;
-	c->order = 0;
-	c->minimized = false;
 	debug("client with pid %d forked\n", c->pid);
 	attach(c);
 	focus(c);
@@ -775,18 +1043,78 @@ create(const char *args[]) {
 
 static void
 copymode(const char *args[]) {
-	if (!sel)
+	if (!sel || sel->editor)
 		return;
-	vt_copymode_enter(sel->term);
-	if (args[0]) {
-		vt_copymode_keypress(sel->term, args[0][0]);
-		draw(sel);
+	if (!(sel->editor = vt_create(sel->h - sel->has_title_line, sel->w, 0)))
+		return;
+
+	char *ed = getenv("DVTM_EDITOR");
+	int *to = &sel->editor_fds[0], *from = NULL;
+	sel->editor_fds[0] = sel->editor_fds[1] = -1;
+
+	if (!ed)
+		ed = getenv("EDITOR");
+	if (!ed)
+		ed = getenv("PAGER");
+	if (!ed)
+		ed = editors[0].name;
+
+	const char **argv = (const char*[]){ ed, "-", NULL, NULL };
+	char argline[32];
+	bool colored = false;
+
+	for (unsigned int i = 0; i < LENGTH(editors); i++) {
+		if (!strcmp(editors[i].name, ed)) {
+			for (int j = 1; editors[i].argv[j]; j++) {
+				if (strstr(editors[i].argv[j], "%d")) {
+					int line = vt_content_start(sel->app);
+					snprintf(argline, sizeof(argline), "+%d", line);
+					argv[j] = argline;
+				} else {
+					argv[j] = editors[i].argv[j];
+				}
+			}
+			if (editors[i].filter)
+				from = &sel->editor_fds[1];
+			colored = editors[i].color;
+			break;
+		}
 	}
+
+	if (vt_forkpty(sel->editor, ed, argv, NULL, NULL, to, from) < 0) {
+		vt_destroy(sel->editor);
+		sel->editor = NULL;
+		return;
+	}
+
+	sel->term = sel->editor;
+
+	if (sel->editor_fds[0] != -1) {
+		char *buf = NULL;
+		size_t len = vt_content_get(sel->app, &buf, colored);
+		char *cur = buf;
+		while (len > 0) {
+			ssize_t res = write(sel->editor_fds[0], cur, len);
+			if (res < 0) {
+				if (errno == EAGAIN || errno == EINTR)
+					continue;
+				break;
+			}
+			cur += res;
+			len -= res;
+		}
+		free(buf);
+		close(sel->editor_fds[0]);
+		sel->editor_fds[0] = -1;
+	}
+
+	if (args[0])
+		vt_write(sel->editor, args[0], strlen(args[0]));
 }
 
 static void
 focusn(const char *args[]) {
-	for (Client *c = clients; c; c = c->next) {
+	for (Client *c = nextvisible(clients); c; c = nextvisible(c->next)) {
 		if (c->order == atoi(args[0])) {
 			focus(c);
 			if (c->minimized)
@@ -798,11 +1126,12 @@ focusn(const char *args[]) {
 
 static void
 focusnext(const char *args[]) {
+	Client *c;
 	if (!sel)
 		return;
-	Client *c = sel->next;
+	for (c = sel->next; c && !isvisible(c); c = c->next);
 	if (!c)
-		c = clients;
+		for (c = clients; c && !isvisible(c); c = c->next);
 	if (c)
 		focus(c);
 }
@@ -813,20 +1142,23 @@ focusnextnm(const char *args[]) {
 		return;
 	Client *c = sel;
 	do {
-		c = c->next;
+		c = nextvisible(c->next);
 		if (!c)
-			c = clients;
+			c = nextvisible(clients);
 	} while (c->minimized && c != sel);
 	focus(c);
 }
 
 static void
 focusprev(const char *args[]) {
+	Client *c;
 	if (!sel)
 		return;
-	Client *c = sel->prev;
-	if (!c)
+	for (c = sel->prev; c && !isvisible(c); c = c->prev);
+	if (!c) {
 		for (c = clients; c && c->next; c = c->next);
+		for (; c && !isvisible(c); c = c->prev);
+	}
 	if (c)
 		focus(c);
 }
@@ -837,10 +1169,12 @@ focusprevnm(const char *args[]) {
 		return;
 	Client *c = sel;
 	do {
-		c = c->prev;
-		if (!c)
+		for (c = c->prev; c && !isvisible(c); c = c->prev);
+		if (!c) {
 			for (c = clients; c && c->next; c = c->next);
-	} while (c->minimized && c != sel);
+			for (; c && !isvisible(c); c = c->prev);
+		}
+	} while (c && c != sel && c->minimized);
 	focus(c);
 }
 
@@ -860,8 +1194,8 @@ killclient(const char *args[]) {
 
 static void
 paste(const char *args[]) {
-	if (sel && copybuf)
-		vt_write(sel->term, copybuf, strlen(copybuf));
+	if (sel && copyreg.data)
+		vt_write(sel->term, copyreg.data, copyreg.len);
 }
 
 static void
@@ -893,7 +1227,7 @@ scrollback(const char *args[]) {
 		vt_scroll(sel->term,  sel->h/2);
 
 	draw(sel);
-	curs_set(vt_cursor(sel->term));
+	curs_set(vt_cursor_visible(sel->term));
 }
 
 static void
@@ -907,19 +1241,39 @@ setlayout(const char *args[]) {
 	unsigned int i;
 
 	if (!args || !args[0]) {
-		if (++layout == &layouts[countof(layouts)])
+		if (++layout == &layouts[LENGTH(layouts)])
 			layout = &layouts[0];
 	} else {
-		for (i = 0; i < countof(layouts); i++)
+		for (i = 0; i < LENGTH(layouts); i++)
 			if (!strcmp(args[0], layouts[i].symbol))
 				break;
-		if (i == countof(layouts))
+		if (i == LENGTH(layouts))
 			return;
 		layout = &layouts[i];
 	}
 	arrange();
 }
 
+static void
+incnmaster(const char *args[]) {
+	int delta;
+
+	if (isarrange(fullscreen) || isarrange(grid))
+		return;
+	/* arg handling, manipulate nmaster */
+	if (args[0] == NULL) {
+		screen.nmaster = NMASTER;
+	} else if (sscanf(args[0], "%d", &delta) == 1) {
+		if (args[0][0] == '+' || args[0][0] == '-')
+			screen.nmaster += delta;
+		else
+			screen.nmaster = delta;
+		if (screen.nmaster < 1)
+			screen.nmaster = 1;
+	}
+	arrange();
+}
+
 static void
 setmfact(const char *args[]) {
 	float delta;
@@ -929,7 +1283,7 @@ setmfact(const char *args[]) {
 	/* arg handling, manipulate mfact */
 	if (args[0] == NULL) {
 		screen.mfact = MFACT;
-	} else if (1 == sscanf(args[0], "%f", &delta)) {
+	} else if (sscanf(args[0], "%f", &delta) == 1) {
 		if (args[0][0] == '+' || args[0][0] == '-')
 			screen.mfact += delta;
 		else
@@ -944,34 +1298,44 @@ setmfact(const char *args[]) {
 
 static void
 startup(const char *args[]) {
-	for (unsigned int i = 0; i < countof(actions); i++)
+	for (unsigned int i = 0; i < LENGTH(actions); i++)
 		actions[i].cmd(actions[i].args);
 }
 
 static void
 togglebar(const char *args[]) {
 	if (bar.pos == BAR_OFF)
-		bar.pos = (BAR_POS == BAR_OFF) ? BAR_TOP : BAR_POS;
+		showbar();
 	else
-		bar.pos = BAR_OFF;
+		hidebar();
+	bar.autohide = false;
 	updatebarpos();
 	redraw(NULL);
 }
 
 static void
-togglebell(const char *args[]) {
-	vt_togglebell(sel->term);
+togglebarpos(const char *args[]) {
+	switch (bar.pos == BAR_OFF ? bar.lastpos : bar.pos) {
+	case BAR_TOP:
+		bar.pos = BAR_BOTTOM;
+		break;
+	case BAR_BOTTOM:
+		bar.pos = BAR_TOP;
+		break;
+	}
+	updatebarpos();
+	redraw(NULL);
 }
 
 static void
 toggleminimize(const char *args[]) {
-	Client *c, *m;
+	Client *c, *m, *t;
 	unsigned int n;
 	if (!sel)
 		return;
 	/* the last window can't be minimized */
 	if (!sel->minimized) {
-		for (n = 0, c = clients; c; c = c->next)
+		for (n = 0, c = nextvisible(clients); c; c = nextvisible(c->next))
 			if (!c->minimized)
 				n++;
 		if (n == 1)
@@ -980,20 +1344,20 @@ toggleminimize(const char *args[]) {
 	sel->minimized = !sel->minimized;
 	m = sel;
 	/* check whether the master client was minimized */
-	if (sel == clients && sel->minimized) {
-		c = sel->next;
+	if (sel == nextvisible(clients) && sel->minimized) {
+		c = nextvisible(sel->next);
 		detach(c);
 		attach(c);
 		focus(c);
 		detach(m);
-		for (; c && c->next && !c->next->minimized; c = c->next);
+		for (; c && (t = nextvisible(c->next)) && !t->minimized; c = t);
 		attachafter(m, c);
 	} else if (m->minimized) {
 		/* non master window got minimized move it above all other
 		 * minimized ones */
 		focusnextnm(NULL);
 		detach(m);
-		for (c = clients; c && c->next && !c->next->minimized; c = c->next);
+		for (c = nextvisible(clients); c && (t = nextvisible(c->next)) && !t->minimized; c = t);
 		attachafter(m, c);
 	} else { /* window is no longer minimized, move it to the master area */
 		vt_dirty(m->term);
@@ -1023,8 +1387,8 @@ zoom(const char *args[]) {
 		return;
 	if (args && args[0])
 		focusn(args);
-	if ((c = sel) == clients)
-		if (!(c = c->next))
+	if ((c = sel) == nextvisible(clients))
+		if (!(c = nextvisible(c->next)))
 			return;
 	detach(c);
 	attach(c);
@@ -1065,7 +1429,7 @@ mouse_zoom(const char *args[]) {
 
 static Cmd *
 get_cmd_by_name(const char *name) {
-	for (unsigned int i = 0; i < countof(commands); i++) {
+	for (unsigned int i = 0; i < LENGTH(commands); i++) {
 		if (!strcmp(name, commands[i].name))
 			return &commands[i];
 	}
@@ -1073,7 +1437,7 @@ get_cmd_by_name(const char *name) {
 }
 
 static void
-handle_cmdfifo() {
+handle_cmdfifo(void) {
 	int r;
 	char *p, *s, cmdbuf[512], c;
 	Cmd *cmd;
@@ -1094,8 +1458,8 @@ handle_cmdfifo() {
 			if (*s && (cmd = get_cmd_by_name(s)) != NULL) {
 				bool quote = false;
 				int argc = 0;
-				/* XXX: initializer assumes MAX_ARGS == 2 use a initialization loop? */
-				const char *args[MAX_ARGS] = { NULL, NULL, NULL}, *arg;
+				const char *args[MAX_ARGS], *arg;
+				memset(args, 0, sizeof(args));
 				/* if arguments were specified in config.h ignore the one given via
 				 * the named pipe and thus skip everything until we find a new line
 				 */
@@ -1168,7 +1532,7 @@ handle_cmdfifo() {
 }
 
 static void
-handle_mouse() {
+handle_mouse(void) {
 #ifdef CONFIG_MOUSE
 	MEVENT event;
 	unsigned int i;
@@ -1183,7 +1547,7 @@ handle_mouse() {
 
 	vt_mouse(msel->term, event.x - msel->x, event.y - msel->y, event.bstate);
 
-	for (i = 0; i < countof(buttons); i++) {
+	for (i = 0; i < LENGTH(buttons); i++) {
 		if (event.bstate & buttons[i].mask)
 			buttons[i].action.cmd(buttons[i].action.args);
 	}
@@ -1193,7 +1557,7 @@ handle_mouse() {
 }
 
 static void
-handle_statusbar() {
+handle_statusbar(void) {
 	char *p;
 	int r;
 	switch (r = read(bar.fd, bar.text, sizeof bar.text - 1)) {
@@ -1216,6 +1580,39 @@ handle_statusbar() {
 	}
 }
 
+static void
+handle_editor(Client *c) {
+	if (!copyreg.data && (copyreg.data = malloc(screen.history)))
+		copyreg.size = screen.history;
+	copyreg.len = 0;
+	while (c->editor_fds[1] != -1 && copyreg.len < copyreg.size) {
+		ssize_t len = read(c->editor_fds[1], copyreg.data + copyreg.len, copyreg.size - copyreg.len);
+		if (len == -1) {
+			if (errno == EINTR)
+				continue;
+			break;
+		}
+		if (len == 0)
+			break;
+		copyreg.len += len;
+		if (copyreg.len == copyreg.size) {
+			copyreg.size *= 2;
+			if (!(copyreg.data = realloc(copyreg.data, copyreg.size))) {
+				copyreg.size = 0;
+				copyreg.len = 0;
+			}
+		}
+	}
+	c->editor_died = false;
+	c->editor_fds[1] = -1;
+	vt_destroy(c->editor);
+	c->editor = NULL;
+	c->term = c->app;
+	vt_dirty(c->term);
+	draw_content(c);
+	wnoutrefresh(c->window);
+}
+
 static int
 open_or_create_fifo(const char *name, const char **name_created) {
 	struct stat info;
@@ -1239,7 +1636,7 @@ open_or_create_fifo(const char *name, const char **name_created) {
 }
 
 static void
-usage() {
+usage(void) {
 	cleanup();
 	eprint("usage: dvtm [-v] [-M] [-m mod] [-d delay] [-h lines] [-t title] "
 	       "[-s status-fifo] [-c cmd-fifo] [cmd...]\n");
@@ -1249,7 +1646,10 @@ usage() {
 static bool
 parse_args(int argc, char *argv[]) {
 	bool init = false;
+	const char *name = argv[0];
 
+	if (name && (name = strrchr(name, '/')))
+		dvtm_name = name + 1;
 	if (!getenv("ESCDELAY"))
 		set_escdelay(100);
 	for (int arg = 1; arg < argc; arg++) {
@@ -1266,7 +1666,7 @@ parse_args(int argc, char *argv[]) {
 			usage();
 		switch (argv[arg][1]) {
 			case 'v':
-				puts("dvtm-"VERSION" © 2007-2014 Marc André Tanner");
+				puts("dvtm-"VERSION" © 2007-2016 Marc André Tanner");
 				exit(EXIT_SUCCESS);
 			case 'M':
 				mouse_events_enabled = !mouse_events_enabled;
@@ -1275,8 +1675,9 @@ parse_args(int argc, char *argv[]) {
 				char *mod = argv[++arg];
 				if (mod[0] == '^' && mod[1])
 					*mod = CTRL(mod[1]);
-				for (unsigned int i = 0; i < countof(keys); i++)
-					keys[i].mod = *mod;
+				for (unsigned int b = 0; b < LENGTH(bindings); b++)
+					if (bindings[b].keys[0] == MOD)
+						bindings[b].keys[0] = *mod;
 				break;
 			}
 			case 'd':
@@ -1313,13 +1714,23 @@ parse_args(int argc, char *argv[]) {
 
 int
 main(int argc, char *argv[]) {
-	int mod = ERR;
+	KeyCombo keys;
+	unsigned int key_index = 0;
+	memset(keys, 0, sizeof(keys));
+	sigset_t emptyset, blockset;
 
+	setenv("DVTM", VERSION, 1);
 	if (!parse_args(argc, argv)) {
 		setup();
 		startup(NULL);
 	}
 
+	sigemptyset(&emptyset);
+	sigemptyset(&blockset);
+	sigaddset(&blockset, SIGWINCH);
+	sigaddset(&blockset, SIGCHLD);
+	sigprocmask(SIG_BLOCK, &blockset, NULL);
+
 	while (running) {
 		int r, nfds = 0;
 		fd_set rd;
@@ -1339,23 +1750,26 @@ main(int argc, char *argv[]) {
 
 		if (bar.fd != -1) {
 			FD_SET(bar.fd, &rd);
-			nfds = max(nfds, bar.fd);
+			nfds = MAX(nfds, bar.fd);
 		}
 
 		for (Client *c = clients; c; ) {
-			if (c->died) {
+			if (c->editor && c->editor_died)
+				handle_editor(c);
+			if (!c->editor && c->died) {
 				Client *t = c->next;
 				destroy(c);
 				c = t;
 				continue;
 			}
-			FD_SET(c->pty, &rd);
-			nfds = max(nfds, c->pty);
+			int pty = c->editor ? vt_pty_get(c->editor) : vt_pty_get(c->app);
+			FD_SET(pty, &rd);
+			nfds = MAX(nfds, pty);
 			c = c->next;
 		}
 
 		doupdate();
-		r = select(nfds + 1, &rd, NULL, NULL, NULL);
+		r = pselect(nfds + 1, &rd, NULL, NULL, NULL, &emptyset);
 
 		if (r == -1 && errno == EINTR)
 			continue;
@@ -1367,22 +1781,24 @@ main(int argc, char *argv[]) {
 
 		if (FD_ISSET(STDIN_FILENO, &rd)) {
 			int code = getch();
-			Key *key;
 			if (code >= 0) {
-				if (mod >= 0) {
-					if ((key = keybinding(mod, code)))
-						key->action.cmd(key->action.args);
-					mod = ERR;
-				} else if (code == KEY_MOUSE) {
+				keys[key_index++] = code;
+				KeyBinding *binding = NULL;
+				if (code == KEY_MOUSE) {
+					key_index = 0;
 					handle_mouse();
-				} else if (is_modifier(code)) {
-					mod = code;
-				} else if ((key = keybinding(ERR, code))) {
-					key->action.cmd(key->action.args);
-				} else if (sel && vt_copymode(sel->term)) {
-					vt_copymode_keypress(sel->term, code);
-					draw(sel);
+				} else if ((binding = keybinding(keys))) {
+					unsigned int key_length = 0;
+					while (key_length < MAX_KEYS && binding->keys[key_length])
+						key_length++;
+					if (key_index == key_length) {
+						binding->action.cmd(binding->action.args);
+						key_index = 0;
+						memset(keys, 0, sizeof(keys));
+					}
 				} else {
+					key_index = 0;
+					memset(keys, 0, sizeof(keys));
 					keypress(code);
 				}
 			}
@@ -1396,26 +1812,26 @@ main(int argc, char *argv[]) {
 		if (bar.fd != -1 && FD_ISSET(bar.fd, &rd))
 			handle_statusbar();
 
-		for (Client *c = clients; c; ) {
-			if (FD_ISSET(c->pty, &rd) && !vt_copymode(c->term)) {
+		for (Client *c = clients; c; c = c->next) {
+			if (FD_ISSET(vt_pty_get(c->term), &rd)) {
 				if (vt_process(c->term) < 0 && errno == EIO) {
-					/* client probably terminated */
-					Client *t = c->next;
-					destroy(c);
-					c = t;
+					if (c->editor)
+						c->editor_died = true;
+					else
+						c->died = true;
 					continue;
 				}
-				if (c != sel && is_content_visible(c)) {
-					draw_content(c);
-					wnoutrefresh(c->window);
-				}
 			}
-			c = c->next;
+
+			if (c != sel && is_content_visible(c)) {
+				draw_content(c);
+				wnoutrefresh(c->window);
+			}
 		}
 
 		if (is_content_visible(sel)) {
 			draw_content(sel);
-			curs_set(vt_cursor(sel->term));
+			curs_set(vt_cursor_visible(sel->term));
 			wnoutrefresh(sel->window);
 		}
 	}
