@@ -1,5 +1,5 @@
 static void fullscreen(void)
 {
-	for (Client *c = clients; c; c = c->next)
+	for (Client *c = nextvisible(clients); c; c = nextvisible(c->next))
 		resize(c, wax, way, waw, wah);
 }
